/*
 * Created by Brett on 30/01/24.
 * Licensed under GNU General Public License V3.0
 * See LICENSE file for license detail
 */
#include "blt/std/assert.h"
#include "blt/std/logging.h"
#include "blt/profiling/profiler_v2.h"
#include <extern.h>
#include <vector>
#include <algorithm>

const std::vector<double> large_c{
        0.139E100, 0.157E100, 0.175E100, 0.256E100, 0.344E100, 0.413E100, 0.503E100, 0.577E100,
        0.614E100, 0.655E100, 0.954E100, 1.392E100, 1.557E100, 1.648E100, 1.690E100, 1.994E100,
        2.174E100, 2.206E100, 3.245E100, 3.510E100, 3.571E100, 4.354E100, 4.980E100, 6.084E100,
        8.351E100};

const std::vector<double> normally_c{-0.4417998157703872,
                                     -0.310841224215176,
                                     -0.2544758389229413,
                                     -0.21509882047762266,
                                     -0.18254731741828356,
                                     -0.1541570107663562,
                                     -0.12852819470327187,
                                     -0.1048199468783084,
                                     -0.08247628697708857,
                                     -0.061100278634889656,
                                     -0.040388574421146996,
                                     -0.020093702456713224,
                                     0.0,
                                     0.020093702456713224,
                                     0.040388574421146996,
                                     0.061100278634889656,
                                     0.08247628697708857,
                                     0.1048199468783084,
                                     0.12852819470327187,
                                     0.1541570107663562,
                                     0.18254731741828356,
                                     0.21509882047762266,
                                     0.2544758389229413,
                                     0.310841224215176,
                                     0.4417998157703872};

const std::vector<double> nearly_c{
        -0.44, -0.31, -0.25, -0.21, -0.18, -0.15, -0.12, -0.10, -0.08, -0.06, -0.04, -0.02,
        0.0, 0.02, 0.04, 0.06, 0.08, 0.10, 0.12, 0.15, 0.18, 0.21, 0.25, 0.31, 0.44};

const auto runs = 500000;

#undef BLT_DEBUG
#define BLT_DEBUG(...)

void shapiro_test_local()
{
    auto normally = normally_c;
    auto nearly = nearly_c;
    auto large = large_c;
    
    std::sort(normally.begin(), normally.end());
    std::sort(nearly.begin(), nearly.end());
    std::sort(large.begin(), large.end());
    
    BLT_START_INTERVAL("shapiro", "local_sorted");
    for (size_t i = 0; i < runs; i++)
    {
        willbruh w1(normally);
        BLT_ASSERT(w1->estimate == 0.9999999999999999);
        BLT_ASSERT(w1->p_value == 1.0);
        BLT_DEBUG("%f // %f", w1->estimate, w1->p_value);
        
        willbruh w2(nearly);
        BLT_ASSERT(w2->estimate == 0.9997987717271388);
        BLT_ASSERT(w2->p_value == 1.0);
        BLT_DEBUG("%f // %f", w2->estimate, w2->p_value);
        
        willbruh w3(large);
        BLT_ASSERT(w3->estimate == 0.8346662753181684);
        BLT_ASSERT(w3->p_value == 0.0009134904817755807);
        BLT_DEBUG("%f // %f", w3->estimate, w3->p_value);
    }
    BLT_END_INTERVAL("shapiro", "local_sorted");
}

void shapiro_test_unsorted()
{
    auto normally = normally_c;
    auto nearly = nearly_c;
    auto large = large_c;
    
    BLT_START_INTERVAL("shapiro", "unsorted_mut");
    for (int i = 0; i < runs; i++)
    {
        willbruh w1(willbert_unsorted(normally.data(), normally.size()));
        BLT_ASSERT(w1->estimate == 0.9999999999999999);
        BLT_ASSERT(w1->p_value == 1.0);
        BLT_DEBUG("%f // %f", w1->estimate, w1->p_value);
        
        willbruh w2(willbert_unsorted(nearly.data(), nearly.size()));
        BLT_ASSERT(w2->estimate == 0.9997987717271388);
        BLT_ASSERT(w2->p_value == 1.0);
        BLT_DEBUG("%f // %f", w2->estimate, w2->p_value);
        
        willbruh w3(willbert_unsorted(large.data(), large.size()));
        BLT_ASSERT(w3->estimate == 0.8346662753181684);
        BLT_ASSERT(w3->p_value == 0.0009134904817755807);
        BLT_DEBUG("%f // %f", w3->estimate, w3->p_value);
    }
    BLT_END_INTERVAL("shapiro", "unsorted_mut");
}

void shapiro_test_copy()
{
    auto normally = normally_c;
    auto nearly = nearly_c;
    auto large = large_c;
    
    BLT_START_INTERVAL("shapiro", "copy");
    for (int i = 0; i < runs; i++)
    {
        willbruh w1(willbert_unsorted_copy(normally.data(), normally.size()));
        BLT_ASSERT(w1->estimate == 0.9999999999999999);
        BLT_ASSERT(w1->p_value == 1.0);
        BLT_DEBUG("%f // %f", w1->estimate, w1->p_value);
        
        willbruh w2(willbert_unsorted_copy(nearly.data(), nearly.size()));
        BLT_ASSERT(w2->estimate == 0.9997987717271388);
        BLT_ASSERT(w2->p_value == 1.0);
        BLT_DEBUG("%f // %f", w2->estimate, w2->p_value);
        
        willbruh w3(willbert_unsorted_copy(large.data(), large.size()));
        BLT_ASSERT(w3->estimate == 0.8346662753181684);
        BLT_ASSERT(w3->p_value == 0.0009134904817755807);
        BLT_DEBUG("%f // %f", w3->estimate, w3->p_value);
    }
    BLT_END_INTERVAL("shapiro", "copy");
}

void shapiro_test_run()
{
    
    shapiro_test_copy();
    shapiro_test_unsorted();
    shapiro_test_local();
    BLT_PRINT_PROFILE("shapiro");
}
